import logging
from sqlalchemy import util
from sqlalchemy.dialects import mssql
from sqlalchemy.dialects.oracle import cx_oracle
from sqlalchemy.dialects.sqlite.pysqlcipher import SQLiteDialect_pysqlcipher
from sqlalchemy.dialects.sqlite.pysqlite import SQLiteDialect_pysqlite
from sqlalchemy.engine.default import DefaultExecutionContext
from sqlalchemy.ext import automap
from sqlalchemy.sql import expression, sqltypes
from sqlalchemy.sql.schema import ColumnDefault
from sqlalchemy.sql.sqltypes import _Binary

from slashdb.models.sqlalchemy_sdb.utils import binary_processor

log = logging.getLogger('slashdb')


def oracle_patches():
    u"""SQLite Oracle monkey patches."""

    def patched_is_many_to_many(automap_base, table):
        return None, None, None
    # patched function because in SlashDB we don't want to skip junction tables.
    automap._is_many_to_many = patched_is_many_to_many

    class _OracleNChar(sqltypes.NCHAR):
        def get_dbapi_type(self, dbapi):
            return dbapi.FIXED_NCHAR

    cx_oracle._OracleNChar = _OracleNChar
    cx_oracle.dialect.colspecs[sqltypes.NCHAR] = _OracleNChar


def patch_mysql_compound_pk():
    u"""Patches for issue #360.

    Code taken from https://bitbucket.org/zzzeek/sqlalchemy/issues/3921

    """

    @util.memoized_property
    @util.dependencies("sqlalchemy.sql.sqltypes")
    def _arg_is_typed(self, sqltypes):
        if self.is_clause_element:
            # need to resolve the lazy impport
            # otherwise we'll get:
            #   ImportError: importlater.resolve_all() hasn't been called (this is sqlalchemy.sql sqltypes)
            sqltypes._resolve()
            return not isinstance(self.arg.type, sqltypes.NullType)
        else:
            return False

    ColumnDefault._arg_is_typed = _arg_is_typed

    def _exec_default(self, column, default, type_):
        if default.is_sequence:
            return self.fire_sequence(default, type_)
        elif default.is_callable:
            self.current_column = column
            return default.arg(self)
        elif default.is_clause_element:
            # TODO: expensive branching here should be
            # pulled into _exec_scalar()
            conn = self.connection
            if not default._arg_is_typed:
                default_arg = expression.type_coerce(default.arg, type_)
            else:
                default_arg = default.arg
            c = expression.select([default_arg]).compile(bind=conn)
            return conn._execute_compiled(c, (), {}).scalar()
        else:
            return default.arg

    DefaultExecutionContext._exec_default = _exec_default


def patch_db2():
    """Fixes #444 problem of duplicated foreign keys caused by table aliases.

    https://github.com/ibmdb/python-ibmdbsa/issues/20
    https://github.com/mdobrzanski/python-ibmdbsa/tree/reflect-with-aliases

    """
    from ibm_db_sa import reflection
    from sqlalchemy import join, sql

    def get_foreign_keys(self, connection, table_name, schema=None, **kw):
        default_schema = self.default_schema_name
        current_schema = self.denormalize_name(schema or default_schema)
        default_schema = self.normalize_name(default_schema)
        table_name = self.denormalize_name(table_name)
        sysfkeys = self.sys_foreignkeys
        systbl = self.sys_tables

        query = sql \
            .select([sysfkeys.c.fkname, sysfkeys.c.fktabschema,
                     sysfkeys.c.fktabname, sysfkeys.c.fkcolname,
                     sysfkeys.c.pkname, sysfkeys.c.pktabschema,
                     sysfkeys.c.pktabname, sysfkeys.c.pkcolname]) \
            .select_from(
            join(systbl, sysfkeys, systbl.c.tabname == sysfkeys.c.pktabname)
        ) \
            .where(systbl.c.type == 'T') \
            .where(systbl.c.tabschema == current_schema) \
            .where(sysfkeys.c.fktabname == table_name) \
            .order_by(systbl.c.tabname)

        fschema = {}
        for r in connection.execute(query):
            if not (r[0]) in fschema:
                referred_schema = self.normalize_name(r[5])

                # if no schema specified and referred schema here is the
                # default, then set to None
                if schema is None and \
                        referred_schema == default_schema:
                    referred_schema = None

                fschema[r[0]] = {
                    'name': self.normalize_name(r[0]),
                    'constrained_columns': [self.normalize_name(r[3])],
                    'referred_schema': referred_schema,
                    'referred_table': self.normalize_name(r[6]),
                    'referred_columns': [self.normalize_name(r[7])]}
            else:
                fschema[r[0]]['constrained_columns'].append(self.normalize_name(r[3]))
                fschema[r[0]]['referred_columns'].append(self.normalize_name(r[7]))
        return [value for key, value in fschema.items()]

    reflection.DB2Reflector.get_foreign_keys = get_foreign_keys


class SQLiteDialectQuotesFixed_pysqlite(SQLiteDialect_pysqlite):
    def __init__(self, *args, **kwargs):
        super(SQLiteDialectQuotesFixed_pysqlite, self).__init__(*args, **kwargs)
        self.identifier_preparer.initial_quote = '['
        self.identifier_preparer.final_quote = ']'


class SQLiteDialectQuotesFixed_pysqlcipher(SQLiteDialect_pysqlcipher):
    def __init__(self, *args, **kwargs):
        super(SQLiteDialectQuotesFixed_pysqlcipher, self).__init__(*args, **kwargs)
        self.identifier_preparer.initial_quote = '['
        self.identifier_preparer.final_quote = ']'


# todo test SQLiteDialect_pysqlcipher
def patch_sqlite_quotes():
    """Fixes unexpected behaviour for SQLite databases when querying for not existing column.

    More detail in #567 and #633

    """

    from sqlalchemy.dialects import registry
    registry.register(
        "sqlite", "slashdb.patches",
        "SQLiteDialectQuotesFixed_pysqlite")

    registry.register(
        "sqlite.pysqlite", "slashdb.patches",
        "SQLiteDialectQuotesFixed_pysqlite")

    registry.register(
        "sqlite.pysqlcipher", "slashdb.patches",
        "SQLiteDialectQuotesFixed_pysqlcipher")


def patch_binary():
    """Patches database binary data types in Data Discovery, #641.

    SQLAlchemy converts binary data into different python type based on database server.
    Since we don't support binary output in current version we override to use Python bytearray as processor.
    When renderer see type bytearray they render text "[binary]" instead of actual value.

    """

    orig_binary_processor = _Binary.result_processor

    def result_processor(self, dialect, coltype):
        return binary_processor

    _Binary.result_processor = result_processor
    mssql.TIMESTAMP.result_processor = orig_binary_processor


def patch_all():
    patches = (
        oracle_patches,
        patch_mysql_compound_pk,
        patch_db2,
        patch_sqlite_quotes,
        patch_binary,
    )

    for patch in patches:
        patch()

    log.debug('Applied patches {}'.format(', '.join(patch.__name__ for patch in patches)))
